package de.xoev.transport.xta;

import com.sun.xml.internal.ws.developer.JAXWSProperties;
import de.xoev.transport.xta._211.ManagementPortType;
import de.xoev.transport.xta._211.MsgBoxPortType;
import de.xoev.transport.xta._211.SendPortType;
import de.xoev.transport.xta._211.XTAService;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;

import javax.net.ssl.KeyManager;
import javax.net.ssl.KeyManagerFactory;
import javax.net.ssl.KeyStoreBuilderParameters;
import javax.net.ssl.SSLContext;
import javax.net.ssl.TrustManager;
import javax.net.ssl.TrustManagerFactory;
import javax.xml.ws.BindingProvider;
import java.io.FileInputStream;
import java.io.IOException;
import java.net.MalformedURLException;
import java.net.URL;
import java.security.InvalidAlgorithmParameterException;
import java.security.KeyManagementException;
import java.security.KeyStore;
import java.security.KeyStoreException;
import java.security.NoSuchAlgorithmException;
import java.security.cert.CertificateException;
import java.util.Map;
import java.util.function.Function;

/*
 * Created by IntelliJ IDEA.
 * User: sweinreuter
 * Date: 22.03.2019
 */
public class XTAServiceManager {

    private final XTAService myXtaService;
    protected final SSLContext mySSLContext;

    protected final TrustManagerFactory myTrustManagerFactory;
    protected final KeyManager myKeyManager;

    private boolean mySchemaValidation;

    public XTAServiceManager(@NotNull final String keystore, @NotNull String password, @NotNull String trustStore, @NotNull String trustPassword, @Nullable Function<String, String> keyPasswordSupplier)
            throws KeyStoreException, NoSuchAlgorithmException, IOException, CertificateException, KeyManagementException, InvalidAlgorithmParameterException
    {
        this(loadKeystore(keystore, password), keyPasswordSupplier != null ? keyPasswordSupplier : a->password);

        init(loadKeystore(trustStore, trustPassword));
    }

    public XTAServiceManager(KeyStore keystore, KeyStore trustStore, @NotNull Function<String, String> keyPasswordSupplier)
            throws KeyStoreException, NoSuchAlgorithmException, KeyManagementException, InvalidAlgorithmParameterException
    {
        this(keystore, keyPasswordSupplier);

        init(trustStore);
    }

    public XTAServiceManager(@NotNull final String keystore, @Nullable String keyAlias, @NotNull String password, @NotNull String trustStore, @NotNull String trustPassword)
            throws KeyStoreException, NoSuchAlgorithmException, IOException, CertificateException, KeyManagementException
    {
        myXtaService = new XTAService();
        mySSLContext = SSLContext.getInstance("SSL");

        myTrustManagerFactory = TrustManagerFactory.getInstance(TrustManagerFactory.getDefaultAlgorithm());
        myKeyManager = new SimpleKeyManager(loadKeystore(keystore, password), keyAlias, password);

        init(loadKeystore(trustStore, trustPassword));
    }

    protected void init(KeyStore trustStore) throws KeyStoreException, KeyManagementException {
        myTrustManagerFactory.init(trustStore);
        mySSLContext.init(new KeyManager[]{ myKeyManager }, getTrustManagers(), null);
    }

    protected XTAServiceManager(KeyStore keyStore, Function<String, String> keyPasswordSupplier) throws NoSuchAlgorithmException, InvalidAlgorithmParameterException {
        myXtaService = new XTAService();
        mySSLContext = SSLContext.getInstance("SSL");

        myTrustManagerFactory = TrustManagerFactory.getInstance(TrustManagerFactory.getDefaultAlgorithm());
        myKeyManager = createKeyManager(keyStore, keyPasswordSupplier);
    }

    static KeyStore loadKeystore(String keystore, String password) throws KeyStoreException, IOException, NoSuchAlgorithmException, CertificateException {
        final KeyStore pkStore = KeyStore.getInstance(keystore.endsWith(".p12") || keystore.endsWith(".pfx") ? "PKCS12" : "JKS");
        pkStore.load(new FileInputStream(keystore), password.toCharArray());
        return pkStore;
    }

    protected TrustManager[] getTrustManagers() {
        return myTrustManagerFactory.getTrustManagers();
    }

    @NotNull
    protected KeyManager createKeyManager(KeyStore keyStore, Function<String, String> keyPasswordSupplier) throws NoSuchAlgorithmException, InvalidAlgorithmParameterException {
        final KeyManagerFactory instance = KeyManagerFactory.getInstance("NewSunX509");
        instance.init(new KeyStoreBuilderParameters(new KeyStore.Builder() {
            @Override
            public KeyStore getKeyStore() {
                return keyStore;
            }

            @Override
            public KeyStore.ProtectionParameter getProtectionParameter(String alias) {
                return new KeyStore.PasswordProtection(keyPasswordSupplier.apply(alias).toCharArray());
            }
        }));
        return instance.getKeyManagers()[0];
    }

    public void enableSchemaValidation() {
        mySchemaValidation = true;
    }

    public ManagementPortType getMgmtPort(final String uri) {
        return preparePort(myXtaService.getManagementPort(), uri);
    }

    public MsgBoxPortType getFetchPort(final String uri) {
        return preparePort(myXtaService.getMsgBoxPort(), uri);
    }

    public SendPortType getSendPort(final String uri) {
        return preparePort(myXtaService.getSendXtaPort(), uri);
    }

    protected <T> T preparePort(T port, String uri) {
        try {
            if (!new URL(uri).getProtocol().equals("https")) {
                throw new IllegalArgumentException("HTTPS-URL expected: " + uri);
            }
        } catch (MalformedURLException e) {
            throw new IllegalArgumentException("Invalid URL: " + uri, e);
        }

        final Map<String, Object> requestContext = ((BindingProvider) port).getRequestContext();
        
        requestContext.put(JAXWSProperties.SSL_SOCKET_FACTORY, mySSLContext.getSocketFactory());
        requestContext.put(BindingProvider.ENDPOINT_ADDRESS_PROPERTY, uri);
        requestContext.put("schema-validation-enabled", String.valueOf(mySchemaValidation));

        return port;
    }
}
